/*
 * Copyright (c) 2010       , Jakub Hrozek <jhrozek@redhat.com>
 * Copyright (c) 2011       , Peter Vrabec <pvrabec@redhat.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the copyright holders or contributors may not be used to
 *    endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>

#ifdef WITH_SELINUX

#include "defines.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdarg.h>
#include <selinux/selinux.h>
#include <semanage/semanage.h>
#include "prototypes.h"
#include "defines.h"

static int safe_system (const char *command,
                 const char *argv[],
                 const char *env[],
                 int ignore_stderr)
{
	int status = -1;
	int fd;
	pid_t pid;
	
	pid = fork();
	if (pid < 0) {
		return -1;
	}

	if (pid) {       /* Parent */
		if (waitpid (pid, &status, 0) > 0) {
			return status;
		} else {
			return -1;
		}
	}

	fd = open ("/dev/null", O_RDWR);
	/* Child */
	dup2 (fd, 0);	// Close Stdin
	if (ignore_stderr) {
		dup2 (fd, 2);	// Close Stderr
	}

	execve (command, (char *const *) argv, (char *const *) env);
	fprintf (stderr, _("Failed to exec '%s'\n"), argv[0]);
	exit (EXIT_FAILURE);
}

static void semanage_error_callback (unused void *varg,
                                     semanage_handle_t *handle,
                                     const char *fmt, ...)
{
	int ret;
	char * message = NULL;
	va_list ap;


	va_start (ap, fmt);
	ret = vasprintf (&message, fmt, ap);
	va_end (ap);
	if (ret < 0) {
		/* ENOMEM */
		return;
	}

	switch (semanage_msg_get_level (handle)) {
	case SEMANAGE_MSG_ERR:
	case SEMANAGE_MSG_WARN:
		fprintf (stderr, _("[libsemanage]: %s\n"), message);
		break;
	case SEMANAGE_MSG_INFO:
		/* nop */
		break;
	}

	free (message);
}


static semanage_handle_t *semanage_init (void)
{
	int ret;
	semanage_handle_t *handle = NULL;

	handle = semanage_handle_create ();
	if (NULL == handle) {
		fprintf (stderr,
		         _("Cannot create SELinux management handle\n"));
		return NULL;
	}

	semanage_msg_set_callback (handle, semanage_error_callback, NULL);

	ret = semanage_is_managed (handle);
	if (ret != 1) {
		fprintf (stderr, _("SELinux policy not managed\n"));
		goto fail;
	}

	ret = semanage_access_check (handle);
	if (ret < SEMANAGE_CAN_READ) {
		fprintf (stderr, _("Cannot read SELinux policy store\n"));
		goto fail;
	}

	ret = semanage_connect (handle);
	if (ret != 0) {
		fprintf (stderr,
		         _("Cannot establish SELinux management connection\n"));
		goto fail;
	}

	return handle;

fail:
	semanage_handle_destroy (handle);
	return NULL;
}


int set_seuser (const char *login_name, const char *seuser_name)
{
	semanage_handle_t *handle = NULL;
	semanage_seuser_key_t *key = NULL;
	int ret;
	int seuser_exists = 0;

	if (NULL == seuser_name) {
		/* don't care, just let system pick the defaults */
		return 0;
	}

	handle = semanage_init ();
	if (NULL == handle) {
		fprintf (stderr, _("Cannot init SELinux management\n"));
		ret = 1;
		goto done;
	}

	ret = semanage_seuser_key_create (handle, login_name, &key);
	if (ret != 0) {
		fprintf (stderr, _("Cannot create SELinux user key\n"));
		ret = 1;
		goto done;
	}

	ret = semanage_seuser_exists (handle, key, &seuser_exists);
	if (ret < 0) {
		fprintf (stderr, _("Cannot verify the SELinux user\n"));
		ret = 1;
		goto done;
	}

	/* Drop libsemanage handles before calling semanage command */
	semanage_seuser_key_free (key);
	semanage_handle_destroy (handle);
	key = NULL;
	handle = NULL;

	if (0 != seuser_exists) {
		const char *argv[7];

		argv[0] = "/usr/sbin/semanage";
		argv[1] = "login";
		argv[2] = "-m";
		argv[3] = "-s";
		argv[4] = seuser_name;
		argv[5] = login_name;
		argv[6] = NULL;
		if (safe_system (argv[0], argv, NULL, 1)) {
			fprintf (stderr,
			         _("Cannot modify SELinux user mapping\n"));
			ret = 1;
			goto done;
		}
	} else {
		const char *argv[7];

		argv[0] = "/usr/sbin/semanage";
		argv[1] = "login";
		argv[2] = "-a";
		argv[3] = "-s";
		argv[4] = seuser_name;
		argv[5] = login_name;
		argv[6] = NULL;
		if (safe_system (argv[0], argv, NULL, 1)) {
			fprintf (stderr,
			         _("Cannot add SELinux user mapping\n"));
			ret = 1;
			goto done;
		}
	}

	ret = 0;

        /* drop obsolete matchpathcon cache */
        matchpathcon_fini();

done:
	semanage_seuser_key_free (key);
	semanage_handle_destroy (handle);
	return ret;
}


int del_seuser (const char *login_name)
{
	const char *args[5];


	args[0] = "/usr/sbin/semanage";
	args[1] = "login";
	args[2] = "-d";
	args[3] = login_name;
	args[4] = NULL;
	safe_system (args[0], args, NULL, 1);

        /* drop obsolete matchpathcon cache */
        matchpathcon_fini();

	return 0;
}
#else				/* !WITH_SELINUX */
extern int errno;		/* warning: ANSI C forbids an empty source file */
#endif				/* !WITH_SELINUX */
